package cz.mendelu.dnaAnalyser.utils;

import cz.mendelu.dnaAnalyser.utils.exception.ImporterException;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.stereotype.Component;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;

import static java.text.MessageFormat.format;

@Slf4j
@Component
public class FirstRunImporter {

    /**
     * Import all files and directories from destination in resources to directory
     * @param srcResource resource directory
     * @param destDir destination directory
     */
    public void importFiles(String srcResource, String destDir) {
        log.debug("Import files from {} to {}.", srcResource, destDir);
        Path destPath = Paths.get(destDir);
        prepareDestDir(destPath);
        List<Resource> resources = findResourcesByMatchingPattern(srcResource + "*");
        resources.forEach(r -> copyResource(r, destPath));
    }

    /**
     * Create Destination dir if not exists.
     * @param destPath
     * @throws ImporterException if something wrong.
     */
    private void prepareDestDir(Path destPath) {
        try {
            if (Files.notExists(destPath)) {
                Files.createDirectories(destPath);
                log.debug("Create destination directory {}", destPath);
            }
        } catch (IOException e) {
            throw new ImporterException(format("Can't create directory {}.", destPath), e);
        }
    }

    private List<Resource> findResourcesByMatchingPattern(String srcPattern) {
        ResourcePatternResolver resourceResolver = new PathMatchingResourcePatternResolver();
        try {
            Resource[] resources = resourceResolver.getResources(srcPattern);
            log.debug("Found {} resources to import from {}", resources.length, srcPattern);
            return Arrays.asList(resources);
        } catch (IOException e) {
            throw new ImporterException(format("Can't found resource by pattern {}.", srcPattern), e);
        }
    }

    private void copyResource(Resource resource, Path destPath) {
        try (InputStream in = resource.getInputStream()) {
            if (in == null) {
                log.error("InputStream from resource {} is null.", resource);
                return;
            }
            Path target = destPath.resolve(resource.getFilename());
            if (Files.exists(target)) {
                log.debug("Skip copy resource from {} to {}.", resource, destPath);
                return;
            }
            try (OutputStream out = new FileOutputStream(target.toFile())) {
                IOUtils.copyLarge(in, out);
                log.info("Extract resource {} to {}", resource, destPath);
            }
        } catch (Exception e) {
            log.warn("Create default resource failed.", e);
        }
    }
//        ry {
//            Path destPath = Paths.get(destDir);
//            if (Files.notExists(destPath)) {
//                Files.createDirectories(destPath);
//            }
//
//            ResourcePatternResolver resourceResolver = new PathMatchingResourcePatternResolver();
//            // Two different location of source directory base on the way, how is app started.
//            Resource[] resources = resourceResolver.getResources("db/data/sequences/*");
//            Arrays.stream(resources)
//                    .forEach(r -> {
//                                try {
//                                    Path target = destPath.resolve(r.getFilename());
//                                    if (Files.notExists(target)) {
//                                        Files.copy(Paths.get(r.getURI()), target);
//                                        log.info("Copy default sequence {}", target);
//                                    } else {
//                                        log.debug("Default sequence {} exists", r.getFilename());
//                                    }
//                                } catch (IOException e) {
//                                    log.warn("Create default sequence {} failed.", r.getFilename(), e);
//                                }
//                            }
//                    );
//
//        } catch (Exception e) {
//            log.warn("Create default sequences failed.", e);
//        }
//    }
}
