package cz.mendelu.dnaAnalyser.sequence.stream

import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide
import spock.lang.Ignore
import spock.lang.Specification

import java.nio.ByteBuffer
import java.nio.file.NoSuchFileException
import java.util.function.Consumer
import java.util.stream.Collectors

class BufferedWindowStreamTest extends Specification {

    // Only for test using foreach, because is foreach method is well tested.
    private List<String> stackPlain(WindowStream stream) {
        def result = []
        stream.forEach { result.add(it.toPlain()) }
        return result
    }

    def "Constructor with FileChannel."() {
        given:
        def sequenceFile = File.createTempFile("ByteBufferWindowStreamTest", ".seq")
        sequenceFile.write("ACTGACTG")

        when:
        def stream = BufferedWindowStream.openFile(sequenceFile.toPath(), 4)

        then:
        stream != null
    }

    def "Constructor with FileChannel on non-exists file"() {
        given:
        def sequenceFile = new File(File.createTempDir(), "missing.seq")

        when:
        BufferedWindowStream.openFile(sequenceFile.toPath(), 4)

        then:
        thrown(NoSuchFileException.class)
    }

    def "forEach(): Sequence = ACTGACTG, windowsSize = 4, basic usage"() {
        given:
        def windowStream = new BufferedWindowStream(ByteBuffer.wrap("ACTGACTG".bytes), 4)
        def stack = []
        Consumer<Window> action = new Consumer<Window>() {
            @Override
            void accept(Window window) {
                stack.add(window.toPlain())
            }
        }

        when:
        windowStream.forEach(action)

        then:
        stack == ["ACTG", "CTGA", "TGAC", "GACT", "ACTG"]
    }

    def "forEach(): Sequence = ACTGACTG, windowsSize = 1, basic usage"() {
        given:
        def windowStream = new BufferedWindowStream(ByteBuffer.wrap("ACTGACTG".bytes), 1)
        def stack = []
        Consumer<Window> action = new Consumer<Window>() {
            @Override
            void accept(Window window) {
                stack.add(window.toPlain())
            }
        }

        when:
        windowStream.forEach(action)

        then:
        stack == ["A", "C", "T", "G", "A", "C", "T", "G"]
    }

    @Ignore
    // FIXME Aktálně stream našte celý soubor do paměti, takže close není potřeba. Ale efektivnější na pamět by bylo číst ji po blocích a tam by to to již mělo smysl.
    def "forEach(): after file is close"() {
        given:
        def sequenceFile = File.createTempFile("ByteBufferWindowStreamTest", ".seq")
        sequenceFile.write("ACTGACTG")

        and:
        def stream = BufferedWindowStream.openFile(sequenceFile.toPath(), 4)
        stream.close()

        when:
        stream.forEach {}

        then:
        thrown(Exception.class)
    }

    def "filter(): Sequence = ACTGACTG, windowsSize = 4, only start with A"() {
        given:
        def windowStream = new BufferedWindowStream(ByteBuffer.wrap("ACTGACTG".bytes), 4)

        when:
        windowStream = windowStream.filter { it.get(0) == Nucleotide.A }

        then:
        def stack = stackPlain(windowStream)
        stack == ["ACTG", "ACTG"]
    }

    def "map(): Sequence = ACTGACTG, windowsSize = 4 using Window.toPlain()"() {
        given:
        def windowStream = new BufferedWindowStream(ByteBuffer.wrap("ACTGACTG".bytes), 4)

        when:
        def streamOfPlain = windowStream.map { it.toPlain() }

        then:
        streamOfPlain.collect(Collectors.toList()) == ["ACTG", "CTGA", "TGAC", "GACT", "ACTG"]
    }

}
